package me.unurled.raxen.config;

import static me.unurled.raxen.utils.Items.listItemStackDeserialize;
import static me.unurled.raxen.utils.Items.mapItemStackDeserialize;
import static me.unurled.raxen.utils.Items.setItemsToInventory;
import static me.unurled.raxen.utils.Utils.debug;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.reflect.TypeToken;
import com.mongodb.client.MongoCollection;
import com.mongodb.client.model.Filters;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import me.unurled.raxen.Raxen;
import me.unurled.raxen.components.entity.Attributes.Attribute;
import me.unurled.raxen.components.entity.player.Inventories;
import me.unurled.raxen.components.entity.player.PlayerUtils;
import me.unurled.raxen.components.entity.player.Storage;
import me.unurled.raxen.components.entity.player.attributes.Attributes;
import me.unurled.raxen.components.entity.player.storages.EnderChest;
import me.unurled.raxen.components.tutorial.Tutorial;
import me.unurled.raxen.manager.entity.PlayerManager;
import me.unurled.raxen.manager.entity.StorageManager;
import me.unurled.raxen.utils.Items;
import me.unurled.raxen.utils.libs.MongoDB;
import me.unurled.raxen.utils.libs.Redis;
import net.kyori.adventure.text.Component;
import org.bson.Document;
import org.bukkit.Bukkit;
import org.bukkit.Material;
import org.bukkit.configuration.file.FileConfiguration;
import org.bukkit.entity.Player;
import org.bukkit.event.inventory.InventoryType;
import org.bukkit.inventory.Inventory;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.PlayerInventory;
import org.bukkit.scheduler.BukkitRunnable;
import org.jetbrains.annotations.NotNull;

public class PlayerConfig {

    private static Raxen main;
    private Redis redis;
    private MongoDB mongoDB;
    private MongoCollection<Document> mongoCollection;
    private static PlayerManager playerManager;
    private static StorageManager sto;
    private Attributes attributes;

    public PlayerConfig(Raxen main) {
        this.main = main;
        sto = main.getManager().getStorageManager();
        this.mongoDB = main.getManager().getLibsManager().getMongoDB();
        this.mongoCollection = mongoDB.getMongoCollection();
        this.redis = main.getManager().getLibsManager().getRedis();
        playerManager = main.getManager().getPlayerManager();
        this.attributes = new Attributes(main);
    }

    /**
     * closes mongo connection
     */
    public void close() {
        mongoDB.close();
    }

    /**
     * save player data to redis
     */
    public void saveUsingRedis(Player p, Attributes attributes, String invStr) {
        Map<String, String> hash = new HashMap<>();
        hash.put("uuid", p.getUniqueId().toString());
        hash.put("name", p.getName());

        hash.put("last_join", String.valueOf(System.currentTimeMillis()));
//        hash.put("play_time", String.valueOf(playerManager.getPlayTime(p)));
        hash.put("location", p.getLocation().toString());

        hash.put("inv", invStr);
        if (playerManager.getRaxenPlayer(p).getStorage() != null) {
            if (
                playerManager.getRaxenPlayer(p).getStorage().getEc() !=
                    null
            ) {
                EnderChest ec = playerManager
                    .getRaxenPlayer(p)
                    .getStorage()
                    .getEc();
                List<String> list = new ArrayList<String>();
                int reverse = 0;
                for (ItemStack it : ec.getEc()) {
                    if (it != null && it.getType() != Material.AIR) {
                        String s = Items.itemStackSerialize(it, reverse);
                        list.add(s);
                    }
                    reverse += 1;
                }
                String ecstr = Items.listItemStackSerialize(list);
                hash.put("enderchest", ecstr);
            }
        }
        HashMap<String, Attribute> attribute = main
            .getManager()
            .getPlayerManager()
            .getAttribute();
        for (String s : attribute.keySet()) {
            if (s.contains("ITEM") || s.contains("MAX")) {
                continue;
            }
            hash.put(s, attribute.get(s).get(p).toString());
        }
        hash.put("level", String.valueOf(playerManager.getRaxenPlayer(p).getLevel()));
        hash.put("exp", String.valueOf(playerManager.getRaxenPlayer(p).getXpForNextLevel()));
        hash.put("kills", String.valueOf(playerManager.getRaxenPlayer(p).getMobKills()));
        hash.put("deaths", String.valueOf(playerManager.getRaxenPlayer(p).getDeaths()));
        hash.put("money", String.valueOf(playerManager.getRaxenPlayer(p).getMoney()));
        // TODO: Save Collections

        Gson gson = new GsonBuilder().create();
        String json = gson.toJson(hash);
        redis.getJedis().set(p.getUniqueId().toString(), json);
    }

    /**
     * save players stuff to mongo db
     *
     * @param player
     * @param attributes
     * @param invstr
     * @param reverse
     */
    public void saveUsingMongoDB(
        @NotNull Player player,
        Attributes attributes,
        String invstr,
        Integer reverse
    ) {
        debug(main, "Saving " + player.getName() + "'s data!");
        Document doc = new Document("uuid", player.getUniqueId().toString())
            .append("name", player.getName())
            .append("inv", invstr);
        HashMap<String, Attribute> attribute = main
            .getManager()
            .getPlayerManager()
            .getAttribute();
        for (String s : attribute.keySet()) {
            debug("saving " + s);
            doc.append(s, attribute.get(s).get(player));
        }
        if (playerManager.getRaxenPlayer(player).getStorage() != null) {
            if (
                playerManager.getRaxenPlayer(player).getStorage().getEc() !=
                    null
            ) {
                debug(main, "Saving " + player.getName() + "'s storage!");
                EnderChest ec = playerManager
                    .getRaxenPlayer(player)
                    .getStorage()
                    .getEc();
                List<String> list = new ArrayList<String>();
                reverse = 0;
                for (ItemStack it : ec.getEc()) {
                    if (it != null && it.getType() != Material.AIR) {
                        String s = Items.itemStackSerialize(it, reverse);
                        list.add(s);
                    }
                    reverse += 1;
                }
                String ecstr = Items.listItemStackSerialize(list);
                doc.append("ec", ecstr);
            }
        }
        Document playerDoc = mongoCollection
            .find(Filters.eq("uuid", player.getUniqueId().toString()))
            .first();
        if (playerDoc == null) {
            mongoCollection.insertOne(doc);
            debug(main, "insert one");
        } else {
            mongoCollection.replaceOne(
                Filters.eq("uuid", player.getUniqueId().toString()),
                doc
            );
            debug(main, "replace one");
        }
    }

    /**
     * save players stuff to yml
     *
     * @param player
     * @param attributes
     * @param invstr
     * @param reverse
     */
    public void saveUsingYml(
        Player player,
        Attributes attributes,
        String invstr,
        Integer reverse
    ) {
        FileConfiguration config = sto.createYml(player);
        config.set("name", player.getName());
        HashMap<String, Attribute> attribute = main
            .getManager()
            .getPlayerManager()
            .getAttribute();
        for (String s : attribute.keySet()) {
            config.set(s, attribute.get(s).get(player));
        }
        config.set("inv", invstr);
        if (playerManager.getRaxenPlayer(player).getStorage() != null) {
            if (
                playerManager.getRaxenPlayer(player).getStorage().getEc() !=
                    null
            ) {
                EnderChest ec = playerManager
                    .getRaxenPlayer(player)
                    .getStorage()
                    .getEc();
                List<String> list = new ArrayList<String>();
                reverse = 0;
                for (ItemStack it : ec.getEc()) {
                    reverse += 1;
                    if (it != null) {
                        String s = Items.itemStackSerialize(it, reverse);
                        list.add(s);
                    }
                }
                String ecstr = Items.listItemStackSerialize(list);
                config.set("ec", ecstr);
            }
        }
        try {
            config.save(
                main.getDataFolder() +
                    "/playerInfo/" +
                    player.getUniqueId() +
                    "/" +
                    "playerInfo.yml"
            );
        } catch (IOException e) {
            e.printStackTrace();
        }
    }


    /**
     * load player stuff from redis
     *
     * @param p player
     */
    public void loadUsingRedis(Player p) {
        String json = redis.getJedis().get(p.getUniqueId().toString());
        Gson gson = new Gson();
        TypeToken<HashMap<String, String>> typeToken = new TypeToken<HashMap<String, String>>() {
        };
        HashMap<String, String> hashMap = gson.fromJson(json, typeToken.getType());
        if (hashMap == null) {
            playerManager.getRaxenPlayer(p).createNewAttribute();
            playerManager.getRaxenPlayer(p).createStorage();
            playerManager.getRaxenPlayer(p).createInventory();
            PlayerUtils.updateSkills(main, p);
            return;
        }
        int maxHealth = 0, health = 0, itemHealth = 0, defense = 0, itemDefense = 0, speed = 0, itemSpeed = 0, strength = 0, itemStrength = 0, maxMana = 0, mana = 0, itemMana = 0, luck = 0, itemLuck = 0;
        try {
            health = Integer.parseInt(hashMap.get("HEALTH"));
            defense = Integer.parseInt(hashMap.get("DEFENSE"));
            speed = Integer.parseInt(hashMap.get("SPEED"));
            strength = Integer.parseInt(hashMap.get("STRENGTH"));
            mana = Integer.parseInt(hashMap.get("MANA"));
            luck = Integer.parseInt(hashMap.get("LUCK"));
        } catch (NumberFormatException e) {
        }
        Attributes.setAttributes(main, p,
            maxHealth,
            health,
            itemHealth,
            defense,
            itemDefense,
            speed,
            itemSpeed,
            strength,
            itemStrength,
            maxMana,
            mana,
            itemMana,
            luck,
            itemLuck
        );
        Inventory ec = Bukkit.createInventory(
            null,
            54,
            Component.text("Ender Chest")
        );
        if (hashMap.get("enderchest") != null) {
            ec =
                setItemsToInventory(
                    mapItemStackDeserialize(hashMap.get("enderchest")),
                    ec
                );
        }
        playerManager.getRaxenPlayer(p).setStorage(ec);
        PlayerInventory inv = p.getInventory();
        inv =
            setItemsToInventory(
                listItemStackDeserialize(hashMap.get("inventory")),
                inv
            );
        p.getInventory().setContents(inv.getContents());
        p.updateInventory();
        me.unurled.raxen.components.entity.player.storages.Inventory invv =
            new me.unurled.raxen.components.entity.player.storages.Inventory(
                inv
            );
        Inventories invvv = new Inventories(invv);
        playerManager.getRaxenPlayer(p).setInventory(inv);
        PlayerUtils.updateSkills(main, p);
    }

    /**
     * load player stuff from mongo db
     *
     * @param player
     */
    public void loadUsingMongoDB(@NotNull Player player) {
        Document playerDoc = mongoCollection
            .find(Filters.eq("uuid", player.getUniqueId().toString()))
            .first();
        if (playerDoc == null) {
            debug(main, "Intiatilising Data for Player: " + player.getName());
            Tutorial tutorial = new Tutorial();
            main.getManager().getTutorialManager().addTutorial(player, tutorial);
            playerManager.getRaxenPlayer(player).createNewAttribute();
            playerManager.getRaxenPlayer(player).createStorage();
            playerManager.getRaxenPlayer(player).createInventory();
            PlayerUtils.updateSkills(main, player);
            return;
        }
        debug(main, "Loading " + player.getName() + "'s data!");
        attributes.setAttributes(
            player,
            playerDoc.getInteger("MAX_HEALTH"),
            playerDoc.getInteger("HEALTH"),
            playerDoc.getInteger("ITEM_HEALTH"),
            playerDoc.getInteger("DEFENSE"),
            playerDoc.getInteger("ITEM_DEFENSE"),
            playerDoc.getInteger("SPEED"),
            playerDoc.getInteger("ITEM_SPEED"),
            playerDoc.getInteger("STRENGTH"),
            playerDoc.getInteger("ITEM_STRENGTH"),
            playerDoc.getInteger("MAX_MANA"),
            playerDoc.getInteger("MANA"),
            playerDoc.getInteger("ITEM_MANA"),
            playerDoc.getInteger("LUCK"),
            playerDoc.getInteger("ITEM_LUCK")
        );
        Inventory ec = Bukkit.createInventory(
            null,
            54,
            Component.text("Ender Chest")
        );
        if (playerDoc.getString("ec") != null) {
            debug(main, "Loading " + player.getName() + "'s storage.");
            ec =
                setItemsToInventory(
                    mapItemStackDeserialize(playerDoc.getString("ec")),
                    ec
                );
        }
        playerManager.getRaxenPlayer(player).setStorage(ec);
        PlayerInventory inv = player.getInventory();
        inv =
            setItemsToInventory(
                listItemStackDeserialize(playerDoc.getString("inv")),
                inv
            );
        player.getInventory().setContents(inv.getContents());
        player.updateInventory();
        me.unurled.raxen.components.entity.player.storages.Inventory invv =
            new me.unurled.raxen.components.entity.player.storages.Inventory(
                inv
            );
        Inventories invvv = new Inventories(invv);
        playerManager.getRaxenPlayer(player).setInventory(inv);
        PlayerUtils.updateSkills(main, player);
    }

    /**
     * load player stuff from yml
     *
     * @param player
     * @throws IOException
     */
    public void loadUsingYml(Player player) throws IOException {
        FileConfiguration config = sto.createYml(player);
        attributes.setAttributes(
            player,
            config.getInt("maxHealth"),
            config.getInt("health"),
            config.getInt("itemHealth"),
            config.getInt("defense"),
            config.getInt("itemDefense"),
            config.getInt("speed"),
            config.getInt("itemSpeed"),
            config.getInt("strength"),
            config.getInt("itemStrength"),
            config.getInt("maxMana"),
            config.getInt("mana"),
            config.getInt("itemMana"),
            config.getInt("luck"),
            config.getInt("itemLuck")
        );
        Inventory ec = Bukkit.createInventory(
            null,
            54,
            Component.text("Ender Chest")
        );
        if ((String) config.get("ec") != null) {
            ec =
                setItemsToInventory(
                    mapItemStackDeserialize((String) config.get("ec")),
                    ec
                );
        }
        EnderChest ecc = new EnderChest(ec);
        Storage storage = new Storage(ecc);
        playerManager.getRaxenPlayer(player).setStorage(ec);
        PlayerInventory inv = (PlayerInventory) Bukkit.createInventory(
            player,
            InventoryType.PLAYER
        );
        inv =
            setItemsToInventory(
                listItemStackDeserialize((String) config.getString("inv")),
                inv
            );
        player.getInventory().setContents(inv.getContents());
        player.updateInventory();
        me.unurled.raxen.components.entity.player.storages.Inventory invv =
            new me.unurled.raxen.components.entity.player.storages.Inventory(
                inv
            );
        Inventories invvv = new Inventories(invv);
        playerManager.getRaxenPlayer(player).setInventory(inv);
    }

    /**
     * saves player stuff
     *
     * @param player
     */
    public void savePlayerConfig(@NotNull Player player) {
        Attributes attributes = new Attributes(main);
        PlayerInventory inv = player.getInventory();

        List<String> listInv = new ArrayList<String>();
        Integer reverse = 0;
        for (ItemStack it : inv) {
            reverse = reverse + 1;
            if (it != null && it.getType() != Material.AIR) {
                String s = Items.itemStackSerialize(it, reverse);
                listInv.add(s);
            }
        }
        String invstr = Items.itemStackSer(inv);
        debug(main, "saving...");
        String storage = sto.getConfig().getString("storage");
        Integer finalReverse = reverse;
        String finalInvstr = invstr;
        new BukkitRunnable() {
            @Override
            public void run() {
                saveUsingMongoDB(player, attributes, finalInvstr, finalReverse);
            }
        }
            .runTaskAsynchronously(main);
        debug(main, "using mongo");
        if (storage == "MYSQL" || storage == "YML") {
            String finalInvstr1 = invstr;
            Integer finalReverse1 = reverse;
            new BukkitRunnable() {
                @Override
                public void run() {
                    saveUsingYml(
                        player,
                        attributes,
                        finalInvstr1,
                        finalReverse1
                    );
                }
            }
                .runTaskAsynchronously(main);
            debug(main, "using yml");
        }

        debug(
            main,
            "Player: " + player.getName() + " data successfully saved!"
        );
    }

    /**
     * load player stuff
     *
     * @param player
     */
    public void loadPlayerConfig(Player player) {
        String st = StorageManager.getConfig().getString("storage");
        String value = "";
        if (st.equals("MONGODB")) {
            new BukkitRunnable() {
                @Override
                public void run() {
                    loadUsingMongoDB(player);
                }
            }
                .runTaskAsynchronously(main);
            value = "mongodb";
        } else {
            value = "yml";
            new BukkitRunnable() {
                @Override
                public void run() {
                    try {
                        loadUsingYml(player);
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                }
            }
                .runTaskAsynchronously(main);
        }
        debug(
            main,
            "Player: " +
                player.getName() +
                " data successfully loaded using " +
                value
        );
    }
}
